#!/usr/python/bin/python
#
# Simple python implementation of 'man', 
Version="0.3 - 19960503"
#   - does not handle "cat"-pages in any way
#   - does not handle "/usr/man/man1.z/file.1" style compressed pages
#   - does not handle "/usr/man/pman/man3" or other nonstandard arch's
#   - does not check for required filters like tbl/eqn
#   - does not have a search order for the sections
# But:
#   - handles compressed and uncompressed pages on same system
#   - has dbm-indexed whatis files like Tom Christiansen's perl-man
#   - handles SO lines if working with whatis file
#   - will display all matching manual pages in sequence
#   - apropos can handle regular expressions.
#
# Installation: create three links: 'man', 'apropos' and 'makewhatis'
#
# Please send any comments by E-mail. I'm a beginner at Python, so
# even hints on programming style are welcome!
#
# (C) Rob W.W. Hooft, 1996 hooft@embl-heidelberg.de
#
# Distribute freely
#
import posix,gdbm,sys,os,string,regex,regsub,getopt

# Which sections do we know?
sections='123456789lno'

# How does the name of a directory under /usr/man look?
Rsubdir=regex.compile('^man['+sections+']$')

# How do we recognize a manual page?
Rmanpage=regex.compile('\\.['+sections+'][A-Za-z]*\\(\\.gz\\)?$')

# How do we recognize a compressed file?
comprext='.gz'
Rcompressed=regex.compile('\\'+comprext+'$')

# How do we recognize the "NAME" line in a manual page?
Rname=regex.compile('^\\.[Ss][Hh] *"?[Nn][Aa][Mm][Ee]"? *"?\\(.*\\)')
Rendname=regex.compile('^\\.[Ss][HhSY]')

# Aliases
Rso=regex.compile('^\\.\\(so\\|SO\\) *\\(.*\\)')

# How do we split names lines?
Rdash=regex.compile(' \\\\?- \\| \\\\(em ')

# Ugly nroff escape sequences and NL to be stripped from whatis keys
Rugly=regex.compile('\\\\f[BRIP1234]')

# Dot commands that are to be ignored in a names line
Rdots=regex.compile('^\\(\\.PP \\|\\.HP \\|\\.I \\|\\.B \\|\\.IX .*\\|\\.SM \\|\\\\}\\|\\.\\\\".*\\)')

# What are the names and default arguments to run these programs?
prog={'zcat':'gzip -dc 2>/dev/null',
      'uncompress':'gzip -df',
      'compress':'gzip -9nf',
      'man':'groff -Tascii -man',
      'pager':'less'}

# Any file smaller than solimit is considered an SO alias only.
solimit=70

# Any compressed file <= comprlimit is considered for uncompression,
# Any uncompressed file > comprlimit is considered for compression
comprlimit=1024

# Exceptions
callerror='man.callerror'
bug='man.bug'
nonameline='man.nonameline'

####################################
# Some I/O routines

def usererror(s):
	sys.stderr.write(s+'\n')
	raise callerror,s

def debug(s):
	#sys.stderr.write(s)
	pass

def warn(s):
	sys.stderr.write(s+'\n')

####################################
# Some General Purpose routines

def OpenFile(filename):
	"""Open file for reading, even if it is compressed"""
	if Rcompressed.search(filename)>=0:
		return os.popen(prog['zcat']+' '+filename,'r')
	else:
		return open(filename)

def SplitName(nameline):
	"""Split NAME line into (names,description)"""
	x=regsub.split(nameline,Rdash)
	if len(x)<2:
		return (None,None)
	y=string.splitfields(x[0],',')
	for i in range(len(y)):
		y[i]=string.strip(y[i])
	return (y,string.join(x[1:]))

def FindName(filename):
	"""Look for the NAME line in the page."""
	input=OpenFile(filename)
	nameline=None
	while 1:
		line=input.readline()
		if line=='':
			input.close()
			raise nonameline
		if Rname.match(line)>=0:
			if len(Rname.group(1))>0:
				nameline=Rname.group(1)
			else:
				nameline=input.readline()
				if nameline[0:4]=='.Nm ':
					line2=input.readline()
					if line2[0:4]=='.Nd ':
						nameline=nameline[4:-1]+' - '+line2[4:]
						break
				nameline=regsub.sub(Rdots,'',nameline[:-1])
			while 1:
				line=input.readline()
				if line=="":
					warn('End of file reading nameline')
					break
				if Rendname.match(line)>=0:
					break
				line=regsub.sub(Rdots,'',line[:-1])
				nameline=nameline+' '+string.strip(line)
			break
	nameline=regsub.gsub(Rugly,'',nameline)
 	input.close()
	return nameline

def CheckCompressedness(filename):
	"""Check whether a file should be uncompressed or compressed.

	if a compressed manual page is smaller than 'comprlimit' bytes,
	it is uncompressed to test whether the uncompressed version is
	bigger than this limit. If the uncompressed version is also small,
	the compressed version will be replaced by the uncompressed version.

	The other way around, if a manpage was uncompressed, but it is
	larger than the limit, it is replaced by a compressed version.

	This routine returns the name of the resulting file."""
	size=os.stat(filename)[6]
	if Rcompressed.search(filename)>=0 and size<=comprlimit:
		infile=OpenFile(filename)
		size=0
		while 1:
			l=len(infile.read(4096))
			if not l:
				break
			size=size+l
		if size<=comprlimit:
			status=os.system(prog['uncompress']+' '+filename)
			if status:
				warn('Could not uncompress '+filename)
			else:
				filename=regsub(Rcompressed,'',filename)
				warn('Uncompressed file '+filename)
	elif Rcompressed.search(filename)<0 and size>comprlimit:
		status=os.system(prog['compress']+' '+filename)
		if status:
			warn('Could not compress file '+filename)
		else:
			filename=filename+comprext
			warn('Compressed file '+filename)
	return filename

def DeleteFile(filename,message):
	try:
		os.unlink(filename)
		warn('Removed '+message+': '+filename)
	except posix.error:
		warn('Could not remove '+message+': '+filename)

def DeleteOldest(file1,file2):
	"""Delete the oldest of two files."""
	mtime1=os.stat(file1)[9]
	mtime2=os.stat(file2)[9]
	if (mtime2>mtime1):
		DeleteFile(file1,'outdated file')
	else:
		DeleteFile(file2,'outdated file')

def MakeAliases(filename,root):
	"""Make SO aliases for all names referenced in the NAME section."""
	if os.stat(filename)[6]<solimit:
		return
	try:
		nameline=FindName(filename)
		names,descr=SplitName(nameline)
		if names is None:
			warn('Names line format error for '+filename)
		else:
			for name in names:
				MakeSO(name,filename,root)
	except nonameline:
		warn('No names line in '+filename)

def MakeSO(name,filename,root):
	"""Make a single SO for 'name' pointing to 'filename'."""
	if string.find(name,' ')>=0:
		warn('Invalid name "'+name+'" for '+filename)
		return
	pointer=filename[len(root)+1:]
	path=os.path.split(filename)[0]
	soname=os.path.join(path,string.strip(name))+'.'+path[-1]
	if soname==filename[0:len(soname)]:
		return
	if os.path.exists(soname):
	        if os.stat(soname)[6]>solimit:
			warn('Big file exists for SO '+soname+' for '+filename)
		return
	if os.path.exists(soname+comprext):
		warn('Compressed file exists for SO '+soname+' for '+filename)
		return
	try:
		sofile=open(soname,'w')
		sofile.write('.SO '+pointer)
		sofile.close()
		warn('Created SO alias '+name+' for '+filename)
	except IOError:
		warn('Could not create SO alias '+soname+' for '+filename)

def CheckSO(filename,root):
	"""Verify that an SO pointer is pointing to an existing file."""
	pointer=ReadSO(filename)
	full=os.path.join(root,pointer)
	if not os.path.exists(full) and not os.path.exists(full+comprext):
		DeleteFile(filename,'pointer to nowhere')

def ReadSO(filename):
	"""Read an SO pointer."""
	input=OpenFile(filename)
	line=input.read()
	input.close()
	if Rso.match(line)>=0:
		return Rso.group(2)
	else:
		return None

def ShowPage(file):
	"""Show a manual page."""
	if Rcompressed.search(file)>=0:
		pipeline=prog['zcat']+' '+file+'|'+prog['man']
	else:
		pipeline=prog['man']+' '+file
	# Only run it through a pager if needed.
	if sys.stdout.isatty():
		pipeline=pipeline+'|'+pager()
	os.system(pipeline)

####################################
# Some Classes

class ManFileLooper:
	"""Loop over all manual pages. 

        If the object is called, the method '_handleManFile' (which is
        abstract in this base class) will be called for each manual
        page. Results are cached, so repeat calls are cheap"""

	def __call__(self):
		try:
			return self.result
		except AttributeError:
			self.result=[]
			self._handleAll()
			return self.result

	def reset(self):
		del self.result

	def _handleAll(self):
		for root in manpath():
			self._handleManRoot(root)

	def _handleManRoot(self,root):
		self.root=root
		for subdir in os.listdir(root):
			if Rsubdir.match(subdir)>=0:
				completedir=os.path.join(root,subdir)
				if os.path.isdir(completedir):
					self._handleManDir(completedir)


	def _handleManDir(self,dir):
		for file in os.listdir(dir):
			pos=Rmanpage.search(file)
			if pos>=0:
				self._handleManFile(os.path.join(dir,file))

	def _handleManFile(self,file):
		raise bug,'Abstract method'

class RestrictedManFileLooper(ManFileLooper):
	"""Same, but search can be restricted to one section of the manual."""
	def __init__(self):
		self.section=None

	def setsection(self,section):
		self.section=section

	def _handleManRoot(self,root):
		for subdir in os.listdir(root):
			if self.section!=None and subdir[3]!=self.section:
				continue
			if Rsubdir.match(subdir)>=0:
				completedir=os.path.join(root,subdir)
				if os.path.isdir(completedir):
					self._handleManDir(completedir)

class StupidManFileFinder(RestrictedManFileLooper):
	"""Return a list of manual pages with a specific name. 

	Does a hard lookup to find them"""
	def __init__(self,name):
		self.name=name
		self.section=None

	def _handleManDir(self,dir):
		for file in os.listdir(dir):
			pos=Rmanpage.search(file)
			if pos==len(self.name) and \
			   string.find(file,self.name)==0:
				self.result.append(os.path.join(dir,file))

class ManFileFinder(StupidManFileFinder):
	"""Smarter version of StupidManFileFinder. 

	Uses a gdbm file if one	exists in the man-roots"""

	def _handleManRoot(self,root):
		try:
			self._handleManRootGdbm(root)
		except gdbm.error:
			StupidManFileFinder._handleManRoot(self,root)

	def _handleManRootGdbm(self,root):
		n=os.path.join(root,'pywhatis')
		index=gdbm.open(n+'.gdbm','r')
		try:
			l=string.splitfields(index[self.name],os.pathsep)
			for file in l:
				if self.section!=None and \
				   file[3]!=self.section:
					continue
				self.result.append(os.path.join(root,file))
		except KeyError:
			pass

class AproposLooper(ManFileLooper):
	"""Search the ASCII whatis files in each manroot."""
	def __init__(self,term):
		self.name=term
		self.R=regex.compile(term)
		if sys.stdout.isatty():
			self.out=os.popen(prog['pager'],'w')
		else:
			self.out=sys.stdout

	def _handleManRoot(self,root):
		n=os.path.join(root,'pywhatis')
		try:
			index=open(n,'r')
		except IOError:
			warn('No whatis file in '+root)
			return
		while 1:
			line=index.readline()
			if line=='':
				break
			if self.R.search(line)>=0:
				self.out.write(line)

class WhatisMaker(ManFileLooper):
	"""Create a whatis file in each manroot."""

	def _handleManRoot(self,root):
		warn('Making whatis in '+root)
		n=os.path.join(root,'pywhatis')
		self.whatisfile=open(n,'w')
		self.whatisdict=gdbm.open(n+'.gdbm','n')
		ManFileLooper._handleManRoot(self,root)
		self.whatisdict.close()
		self.whatisfile.close()

	def _handleManFile(self,filename):
		if os.path.islink(filename):
			# just assume we'll see it under the other name
			return

		# Construct the pointer for the dbm index
		subfile=filename[len(self.root)+1:]

		# Construct a standard name for the dbm index
		head,thisname=os.path.split(filename)
		thisname=regsub.sub(Rcompressed,'',thisname)
		thisname,ext=os.path.splitext(thisname)

		if os.stat(filename)[6]<solimit:
			pointer=ReadSO(filename)
			if pointer is None:
				warn('Small file not SO alias: '+filename)
			else:
				self._handleso(thisname,pointer)
			return
		try:
			nameline=FindName(filename)
		except nonameline:
			warn('No names line in '+filename)
			# save what can be saved....
			self._addkey(thisname,subfile)
			return
		# add the name itself to the index
		self._addkey(thisname,subfile)
		# Put chapter number in nameline, and write it to whatis file
		self.whatisfile.write(
			regsub.sub(Rdash,' ('+ext[1:]+') - ',nameline)+'\n')
		debug('NAME '+filename+' '+nameline)
		# add an index key for each of the names
		names,descr=SplitName(nameline)
		if names==None:
			return
		for name in names:
			if name!=thisname:
				self._addkey(name,subfile)

	def _addkey(self,name,subfile):
		"""Add key to the index. No duplicates!!"""
		# make the key acceptable
		name=string.lower(name)
		debug('KEY '+name+' '+subfile)
		# look it up and add reference
		if self.whatisdict.has_key(name):
			it=self.whatisdict[name]
			if string.find(it,subfile)<0:
				self.whatisdict[name]=it+os.pathsep+subfile
		else:
			self.whatisdict[name]=subfile

	def _handleso(self,thisname,soname):
		"""See if we can find the SO file, and add it to the index."""
		xname=os.path.join(self.root,soname)
		if os.path.exists(xname):
			self._addkey(thisname,soname)
			return
		xname=xname+comprext
		if os.path.exists(xname):
			self._addkey(thisname,soname+comprext)
		else:
			warn('SO file not found "%s"\n'%\
					 os.path.join(self.root,soname))

class Maintainance(ManFileLooper):
	def _handleManFile(self,filename):
		if os.path.islink(filename):
			DeleteFile(filename,'symbolic link')
			return
		if os.path.isdir(filename):
			warn('Unexpected directory '+filename)
			return
		# Multiple links to one file not desired: use SOalias instead
		nlink=os.stat(filename)[3]
		if nlink>1:
			DeleteFile(filename,'multiply linked file')
			return
		# Delete oldest from a compressed/uncompressed pair
		if os.path.exists(filename+comprext):
			DeleteOldest(filename,filename+comprext)
			if not os.path.exists(filename):
				return
		# Small files should be uncompressed, large files compressed.
		filename=CheckCompressedness(filename)
		if os.stat(filename)[6]>=solimit:
			# For big files, make their SO aliases
			MakeAliases(filename,self.root)
		else:
			# For small files, see if they are not SO-ing nowhere
			CheckSO(filename,self.root)
		
class CachedReadonlyVariable:
	"""Class that can be used instead of a readonly global variable.

	If a global variable is expensive to initialize and not always
        needed, this class can be used to do the work. The
        initialization will be performed the first time the __call__
        method is used. All further invocations are 'cheap'. These things
	are obviously meant to be instantiated only once...."""

	def __call__(self):
		try:
			return self.value
		except AttributeError:
			self._setup()
			return self.value

	def Set(self,value):
		"""Dirty hack to be able to overwrite the value"""
		self.value=value

	def _setup(self):
		raise bug,'Abstract method'

class ManPath(CachedReadonlyVariable):
	"""Get the MANPATH.

	Results are based either on the environment variable MANPATH
	or, if that one does not exist, on the PATH variable."""
	def _setup(self):
		try:
			MANPATH=os.environ['MANPATH']
			self.value=string.splitfields(MANPATH,os.pathsep)
			return
		except KeyError:
			try:
				PATH=os.environ['PATH']
				path=string.splitfields(PATH,os.pathsep)
			except KeyError:
				self.value=['/usr/man','/usr/local/man']
				return
		Rbin=regex.compile(os.sep+'bin$')
		self.value=[]
		mandict={}
		for dir in path:
			mandir=regsub.sub(Rbin,os.sep+'man',dir)
			if dir!=mandir and os.path.isdir(mandir):
				loop=0
				while os.path.islink(mandir) and loop<10:
					mandir=os.readlink(mandir)
					loop=loop+1
				if not mandict.has_key(mandir):
					mandict[mandir]=1
					self.value.append(mandir)

class Pager(CachedReadonlyVariable):
	"""Find the PAGER to use."""
	def _setup(self):
		try:
			self.value=os.environ['MANPAGER']
			return
		except KeyError:
			pass
		try:
			self.value=os.environ['PAGER']
			return
		except KeyError:
			pass
	       	self.value=prog['pager']

####################################
# The main programs for the different modes

def man(options,arguments):
	"""Driver for the 'man' command itself."""
	if options.has_key('-l'):
		if len(arguments)==1:
			if os.path.exists(arguments[0]):
				ShowPage(arguments[0])
			else:
				usererror('Cannot open file '+arguments[0])
		else:
			usererror('Wrong number of arguments')
		return
	if len(arguments)==1:
		finder=ManFileFinder(arguments[0])
	elif len(arguments)==2:
		if not string.find(sections,arguments[0])>=0:
			usererror('Unknown section "%s"'%arguments[0])
		finder=ManFileFinder(arguments[1])
		finder.setsection(arguments[0])
	else:
		usererror('Wrong number of arguments')

	if len(finder())==0:
		sys.stderr.write('No manual entry for '+arguments[-1]+'\n')
		return
	for file in finder():
		ShowPage(file)

def apropos(options,arguments):
	"""Driver for the 'apropos' command."""
	if not len(arguments)==1:
		usererror('Wrong number of arguments')
	looper=AproposLooper(arguments[0])
	looper()

def makewhatis(options,arguments):
	"""Driver for the 'makewhatis' command."""
	if len(arguments)==1:
		manpath.Set([arguments[0]])
	elif len(arguments)==0:
		pass
	else:
		usererror('Wrong number of arguments')
	looper=WhatisMaker()
	looper()

def maintain(options,arguments):
	"""Driver for manpage maintainance."""
	if len(arguments)==1:
		manpath.Set([arguments[0]])
	elif len(arguments)==0:
		pass
	else:
		usererror('Wrong number of arguments')
	m=Maintainance()
	m()

def usage():
	sys.stderr.write("""Usage:
	man [section] page    : View manual pages
	man -v                : display version number
	man -l filename       : View manual page by filename
	man -k keyword        : List appropriate manual pages
	man -m manroot        : Perform maintainance [CAREFUL!]
	                        (needs write access to man root)
	apropos keyword       : idem.
	makewhatis manroot    : Brings 'pywhatis' files up to date
                                (needs write access to man root)
""")

def main():
	command={'man':man,
		 'apropos':apropos,
		 'makewhatis':makewhatis,
		 'maintain':maintain}

	progpath,mode=os.path.split(sys.argv[0])

	if mode=='man':
		try:
			optionlist,arguments=getopt.getopt(sys.argv[1:],'klvm')
		except getopt.error:
			usererror('Unrecognized option')
	else:
		try:
			optionlist,arguments=getopt.getopt(sys.argv[1:],'')
		except getopt.error:
			usererror('Unrecognized option')

	options={}
	for option in optionlist:
		options[option[0]]=option[1]

	if options.has_key('-k'):
		mode='apropos'
		del options['-k']

	if options.has_key('-m'):
		if not mode=='man':
			usererror('Incompatible flags')
		mode='maintain'
		del options['-m']

	if options.has_key('-v'):
		print "Python implementation of man, Version "+Version
		return

	try:
		command[mode](options,arguments)
	except KeyError:
		usererror('Unrecognized program name: '+sys.argv[0])

manpath=ManPath()
pager=Pager()

try:
	main()
except callerror:
	usage()
except KeyboardInterrupt:
	pass
except IOError:
	pass
